/*
    libmaus2
    Copyright (C) 2009-2013 German Tischler
    Copyright (C) 2011-2013 Genome Research Limited

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#if ! defined(LIBMAUS2_SOCKET_HPP)
#define LIBMAUS2_SOCKET_HPP

#if defined(LIBMAUAS2_HAVE_UNISTD_H)
#include <unistd.h>
#endif

#if defined(LIBMAUS2_HAVE_SYS_TYPES_H)
#include <sys/types.h>
#endif

#if defined(LIBMAUS2_HAVE_SYS_STAT_H)
#include <sys/stat.h>
#endif

#if defined(LIBMAUS2_HAVE_SYS_SOCKET_H)
#include <sys/socket.h>
#endif

#if defined(LIBMAUS2_HAVE_NETINET_IN_H)
#include <netinet/in.h>
#endif

#if defined(LIBMAUS2_HAVE_FCNTL_H)
#include <fcntl.h>
#endif

#if defined(LIBMAUS2_HAVE_SYS_IOCTL_H)
#include <sys/ioctl.h>
#endif

#include <libgen.h>

#if defined(LIBMAUS2_HAVE_NETDB_H)
#include <netdb.h>
#endif

#if defined(LIBMAUS2_HAVE_ARPA_INET_H)
#include <arpa/inet.h>
#endif

#if defined(LIBMAUS2_HAVE_NETINET_TCP_H)
#include <netinet/tcp.h>
#endif

#if defined(LIBMAUS2_HAVE_SYS_RESOURCE_H)
#include <sys/resource.h>
#endif

#include <libmaus2/LibMausWindows.hpp>

#include <climits>
#include <cstdlib>
#include <cstring>
#include <cerrno>
#include <iostream>
#include <set>

#include <libmaus2/exception/LibMausException.hpp>
#include <libmaus2/util/unique_ptr.hpp>
#include <libmaus2/util/shared_ptr.hpp>
#include <libmaus2/network/GetHostName.hpp>
#include <libmaus2/network/SocketInputOutputInterface.hpp>
#include <stdexcept>

#include <libmaus2/posix/PosixFunctions.hpp>

#if defined(__APPLE__)
#include <netinet/tcp.h>
#endif

namespace libmaus2
{
	namespace network
	{
		struct SocketBase : public SocketInputOutputInterface
		{
			public:
			typedef SocketBase this_type;
			typedef ::libmaus2::util::unique_ptr<this_type>::type unique_ptr_type;
			typedef ::libmaus2::util::shared_ptr<this_type>::type shared_ptr_type;

			private:
			int fd;

			protected:
			struct sockaddr_in remaddr;
			bool remaddrset;

			private:
			void cleanup()
			{
				if ( fd != -1 )
				{
					close(fd);
					fd = -1;
				}
			}

			std::string getStringAddr()
			{
				if ( remaddrset )
				{
					std::ostringstream ostr;
					uint32_t const rem = ntohl(remaddr.sin_addr.s_addr);
					ostr
						<< ((rem >> 24) & 0xFF) << "."
						<< ((rem >> 16) & 0xFF) << "."
						<< ((rem >>  8) & 0xFF) << "."
						<< ((rem >>  0) & 0xFF);
					return ostr.str();
				}
				else
				{
					return "<unknown addr>";
				}
			}

			public:
			void write(char const * data, size_t len)
			{
				while ( len )
				{
					::libmaus2::ssize_t wr = ::libmaus2::posix::PosixFunctions::write ( fd, data, len );

					if ( wr < 0 )
					{
						if ( errno == EINTR )
						{
							std::cerr << "write interrupted by signal." << std::endl;
						}
						else
						{
							::libmaus2::exception::LibMausException se;
							se.getStream() << "SocketBase::write() to " << getStringAddr() << " failed: " << strerror(errno);
							se.finish();
							throw se;
						}
					}

					data += wr;
					len -= wr;
				}
			}

			::libmaus2::ssize_t read(char * data, size_t len)
			{
				::libmaus2::ssize_t totalred = 0;

				while ( len )
				{
					::libmaus2::ssize_t red = ::libmaus2::posix::PosixFunctions::read(fd,data,len);

					if ( red > 0 )
					{
						totalred += red;
						data += red;
						len -= red;
					}
					else if ( red < 0 && errno == EINTR )
					{
						std::cerr << "read interrupted by signal." << std::endl;
					}
					else
					{
						len = 0;
					}
				}

				return totalred;
			}

			::libmaus2::ssize_t readPart(char * data, size_t len)
			{
				::libmaus2::ssize_t totalred = 0;

				while ( (! totalred) && len )
				{
					::libmaus2::ssize_t red = ::libmaus2::posix::PosixFunctions::read(fd,data,len);

					if ( red > 0 )
					{
						totalred += red;
						data += red;
						len -= red;
					}
					else if ( red < 0 && errno == EINTR )
					{
						std::cerr << "read interrupted by signal." << std::endl;
					}
					else
					{
						len = 0;
					}
				}

				return totalred;
			}

			void setKeepAlive(
				int const
					#if defined(LIBMAUS2_HAVE_SETSOCKOPT)
					enable
					#endif
					, // boolean
				int const
					#if defined(__linux__) || defined(__APPLE__)
					timeout // time until first check/ping (seconds)
					#endif
					,
				int const
					#if defined(__linux__)
					num
					#endif
					// number of checks
					,
				int const
					#if defined(__linux__)
					intv
					#endif
					// interval between checks/pings (seconds)
			)
			{
				#if defined(LIBMAUS2_HAVE_SETSOCKOPT) && ! defined(_WIN32)
				setsockopt(getFD(), SOL_SOCKET, SO_KEEPALIVE, &enable, sizeof(enable));

				#if defined(__linux__)
				setsockopt(getFD(), IPPROTO_TCP, TCP_KEEPIDLE, &timeout, sizeof(timeout));
				#elif defined(__APPLE__)
				setsockopt(getFD(), IPPROTO_TCP, TCP_KEEPALIVE, &timeout, sizeof(timeout));
				#endif

				#if defined(__linux__)
				setsockopt(getFD(), IPPROTO_TCP, TCP_KEEPCNT, &num, sizeof(num));
				setsockopt(getFD(), IPPROTO_TCP, TCP_KEEPINTVL, &intv, sizeof(intv));
				#endif
				#endif
			}

			protected:
			static void setAddress(char const * hostname, sockaddr_in & recadr)
			{
				if ( hostname )
				{
					#if defined(LIBMAUS2_HAVE_GETHOSTBYNAME2)
					struct hostent * he = gethostbyname2(hostname,AF_INET);
					#elif defined(LIBMAUS2_HAVE_GETHOSTBYNAME)
					struct hostent * he = gethostbyname(hostname);
					#else
					#error "Neither gethostbyname2 nor gethostbyname available"
					#endif

					if ( ! he )
					{
						::libmaus2::exception::LibMausException se;
						#if defined(LIBMAUS2_HAVE_HSTRERROR)
						se.getStream() << "failed to get address for " << hostname << " via gethostbyname: " << hstrerror(h_errno);
						#else
						se.getStream() << "failed to get address for " << hostname << " via gethostbyname: " << h_errno;
						#endif
						se.finish();
						throw se;
					}

					recadr.sin_family = he->h_addrtype;

					if ( he->h_addr_list[0] == 0 )
					{
						::libmaus2::exception::LibMausException se;
						se.getStream() << "failed to get address for " << hostname << " via gethostbyname (no address returned)";
						se.finish();
						throw se;
					}
					else
					{
						if ( recadr.sin_family == AF_INET )
						{
							memcpy ( &recadr.sin_addr.s_addr, he->h_addr_list[0], he->h_length );
						}
						else
						{
							::libmaus2::exception::LibMausException se;
							se.getStream() << "only IPv4 supported";
							se.finish();
							throw se;
						}

						#if 0
						if ( recadr.sin_family == AF_INET )
						{
							std::cerr << inet_ntoa(recadr.sin_addr) << std::endl;
						}
						#endif
					}
				}
				else
				{
					recadr.sin_family = AF_INET;
					recadr.sin_addr.s_addr = INADDR_ANY;
				}
			}

			public:
			SocketBase()
			: fd(-1), remaddrset(false)
			{
				fd = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);

				if ( fd < 0 )
				{
					cleanup();
					::libmaus2::exception::LibMausException se;
					se.getStream() << "socket() failed: " << strerror(errno);
					se.finish();
					throw se;
				}
			}

			SocketBase(int rfd, sockaddr_in const * aadr = 0)
			: fd(rfd), remaddrset(aadr != 0)
			{
				if ( aadr )
				{
					memcpy ( &remaddr, aadr, sizeof(sockaddr_in) );
				}
			}

			~SocketBase()
			{
				cleanup();
			}

			bool hasData()
			{
				fd_set readfds;
				FD_ZERO(&readfds);
				FD_SET(fd,&readfds);
				struct timeval tv = { 0,0 };
				int selok = ::select(fd+1,&readfds,0,0,&tv);
				return selok > 0;
			}

			int getFD() const
			{
				return fd;
			}

			int releaseFD()
			{
				int const rfd = fd;
				fd = -1;
				return rfd;
			}

		};

		struct ClientSocket : public SocketBase
		{
			typedef ClientSocket this_type;
			typedef ::libmaus2::util::unique_ptr<this_type>::type unique_ptr_type;

			ClientSocket(unsigned short port, char const * hostname)
			: SocketBase()
			{
				setAddress(hostname, remaddr);
				remaddr.sin_port = htons(port);
				remaddrset = true;

				if ( connect(getFD(),reinterpret_cast<const sockaddr *>(&remaddr),sizeof(remaddr)) != 0 )
				{
					::libmaus2::exception::LibMausException se;
					se.getStream() << "connect() failed: " << strerror(errno);
					se.finish();
					throw se;
				}
			}

			static SocketBase::unique_ptr_type baseCast(ClientSocket::unique_ptr_type & C)
			{
				int const fd = C->getFD();
				C->releaseFD();

				try
				{
					SocketBase::unique_ptr_type ptr ( new SocketBase(fd) );
					return ptr;
				}
				catch(...)
				{
					close(fd);
					throw;
				}
			}

			static SocketBase::unique_ptr_type baseAlloc(unsigned short const port, char const * const hostname)
			{
				ClientSocket::unique_ptr_type ptr ( new ClientSocket(port,hostname) );
				SocketBase::unique_ptr_type baseptr = baseCast(ptr);
				return baseptr;
			}

			static SocketBase::unique_ptr_type baseAlloc(unsigned short const port, std::string const & hostname)
			{

				SocketBase::unique_ptr_type baseptr = baseAlloc(port,hostname.c_str());
				return baseptr;
			}
		};

		struct ServerSocket : public SocketBase
		{
			typedef ServerSocket this_type;
			typedef ::libmaus2::util::unique_ptr<this_type>::type unique_ptr_type;

			sockaddr_in recadr;

			unsigned short getPort() const
			{
				return ntohs(recadr.sin_port);
			}

			static unique_ptr_type allocateServerSocket(
				unsigned short & port,
				unsigned int const backlog,
				std::string const & hostname,
				unsigned int tries)
			{
				unique_ptr_type ptr(allocateServerSocket(port,backlog,hostname.c_str(),tries));
				return ptr;
			}

			static unique_ptr_type allocateServerSocket(
				unsigned short & port,
				unsigned int const backlog,
				char const * hostname,
				unsigned int tries)
			{
				for ( unsigned int i = 0; i < tries; ++i )
				{
					try
					{
						return unique_ptr_type ( new this_type(port,backlog,hostname) );
					}
					catch(std::exception const & ex)
					{
						// std::cerr << ex.what() << std::endl;
						port++;
					}
				}

				::libmaus2::exception::LibMausException ex;
				ex.getStream() << "Failed to allocate ServerSocket (no ports available)";
				ex.finish();
				throw ex;
			}

			ServerSocket(unsigned short rport, unsigned int backlog, char const * hostname)
			: SocketBase()
			{
				memset(&recadr,0,sizeof(recadr));

				setAddress(hostname, recadr);
				recadr.sin_port = htons(rport);

				if ( bind ( getFD(), reinterpret_cast<struct sockaddr *>(&recadr), sizeof(recadr) ) != 0 )
				{
					if ( errno == EADDRINUSE )
						throw std::runtime_error("bind: address is already in use.");

					// cleanup();
					::libmaus2::exception::LibMausException se;
					se.getStream() << "bind() failed: " << strerror(errno);
					se.finish();
					throw se;
				}

				if ( listen ( getFD(), backlog ) != 0 )
				{
					// cleanup();
					::libmaus2::exception::LibMausException se;
					se.getStream() << "listen() failed: " << strerror(errno);
					se.finish();
					throw se;
				}
			}

			~ServerSocket()
			{
			}

			SocketBase::unique_ptr_type accept()
			{
				sockaddr_in aadr;

				#if defined(LIBMAUS2_HAVE_SOCKLEN_T)
				socklen_t
				#else
				int
				#endif
					len = sizeof(aadr);

				int afd = ::accept(getFD(),reinterpret_cast<struct sockaddr *>(&aadr),&len);

				if ( afd < 0 )
				{
					::libmaus2::exception::LibMausException se;
					se.getStream() << "accept() failed: " << strerror(errno);
					se.finish();
					throw se;
				}
				else
				{
					return SocketBase::unique_ptr_type(new SocketBase(afd,&aadr));
				}
			}

			SocketBase::shared_ptr_type acceptShared()
			{
				SocketBase::unique_ptr_type uptr(accept());
				SocketBase * ptr = uptr.release();
				SocketBase::shared_ptr_type sptr(ptr);
				return sptr;
			}

			bool waitForConnection(unsigned int const t)
			{
				fd_set fds;
				FD_ZERO(&fds);
				FD_SET(getFD(),&fds);
				struct timeval timeout = {
					static_cast<long>(t),
					static_cast<long>(0)
				};

				int const r = ::select(getFD()+1,&fds,0,0,&timeout);

				if ( r < 0 )
				{
					::libmaus2::exception::LibMausException se;
					se.getStream() << "select failed on socket." << std::endl;
					se.finish();
					throw se;
				}
				else if ( r == 0 )
				{
					return false;
				}
				else
				{
					return true;
				}
			}
		};
	}
}
#endif
